/*

GooBer Portable Gameboy Emulator
Copyright (C) 2002 Ian Smith

This Program Is Free Software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdlib.h>
#include <string.h>

#include "allegro.h"
#include "gbcpu.h"
#include "lcdc.h"
#include "mem.h"
#include "misc.h"
#include "intr.h"

#define LYC_INTERRUPT     64
#define OAM_INTERRUPT     32
#define VBLANK_INTERRUPT  16
#define HBLANK_INTERRUPT   8
#define LYC_COINCEDENCE    4

#define LCD_DISPLAY     128
#define WND_USE_BGMAP1  64
#define DISPLAY_WND     32
#define USE_TILE0       16
#define USE_BGMAP1      8
#define SPRITE16        4
#define DISP_SPRITES    2
#define DISPLAY_BG      1
 
#define PATTERN0  0
#define PATTERN1  0x800

#define BACKGROUND0  0x1800
#define BACKGROUND1  0x1C00
  
#define CACHE_SIZE 0x1000

#define Y_FLIP 64
#define X_FLIP 32
#define USE_OBP1 16

/* Holds the buffer to render to */
BITMAP *lcdc_buffer;

/* Holds the frame count */
unsigned int lcdc_frames=0;

/* Flags when to redraw the LCDC display */
int lcdc_refresh=0;

/* Temporary buffer that lcdc_render_scanline() renders to */
static BITMAP *lcdc_screen_buffer;

/* Primitive tile cache used by get_tile_bitmap, points to decoded tiles in the
   tile list */
static void *tile_cache[CACHE_SIZE]={NULL};

/* I/O Registers for LCD display */
static unsigned char lcdc_control=0; /* FF40 */
static unsigned char lcdc_status=0;  /* FF41 */
static unsigned char scroll_x=0;     /* FF43 */
static unsigned char scroll_y=0;     /* FF42 */
static unsigned char lcdc_y=0;       /* FF44 */
static unsigned char lcdc_lyc=0;     /* FF45 */
static unsigned char dma=0;          /* FF46 */
static unsigned char bgp=0;          /* FF47 */
static unsigned char obp0=0;         /* FF48 */
static unsigned char obp1=0;         /* FF49 */
static unsigned char window_x=0;     /* FF4A */
static unsigned char window_y=0;     /* FF4B */

/* List of sprites */
static struct 
{
  int x, y;
  int use_obp1;
  unsigned char lower_tile[64], upper_tile[64];  
} sprites[40]={ 0, 0, 0, {0}, {0} };

/* Line buffer, extra speed boost i guess */
static BITMAP *linebuffer;

/* Sprites image */
static BITMAP *sprite_img;

/* Palette */
static RGB white = { 0xFF/4, 0xE0/4, 0x8F/4 };
static RGB light = { 0xDF/4, 0xB0/4, 0x5F/4 };
static RGB dark  = { 0xAE/4, 0x7F/4, 0x3F/4 };
static RGB black = { 0x1F/4, 0x3F/4, 0x4F/4 };

/* Shades of gray for GB palette */
/* static int shades[] = { 248, 249, 250, 251 }; */
static int shades[] = { 16, 17, 18, 19 };

/* Byte-per-shade palette reflections of BGP, OBP0, OBP1 */
static unsigned char bg_pal[4] = {0};
static unsigned char ob_pal0[4] = {0};
static unsigned char ob_pal1[4] = {0};

/* Current phase information */
static unsigned char current_mode;
static unsigned char next_mode;
static unsigned long mode_cycles;
static unsigned long cycle_count;

/* VRAM buffer */
static unsigned char vram[0x2000]={0};

/* OAM buffer */
static unsigned char oam[0x9F]={0};

/* Pointer to the Background map used by the Window */
unsigned char *wndmap = &(vram[0x1800]);
unsigned char *bgmap = &(vram[0x1800]);

/* X/Y offset to draw the LCDC screen buffer to */
static int screen_x = 0;
static int screen_y = 0;

/* Width/Height to scale LCDC display to */
static int scaled=0;
static int screen_w = 160;
static int screen_h = 144;

/* Internal function declarations --------------------------------------------------------- */

/* Memory access handlers */
static unsigned char vram_read(unsigned short addr);
static void vram_write(unsigned short addr, unsigned char value);

static unsigned char oam_read(unsigned short addr);
static void oam_write(unsigned short addr, unsigned char value);

/* Convert an OAM entry into a structure in the sprite list */
static void load_sprite(unsigned char index);

/* Decode a tile */
static void decode_tile(unsigned char tile[], unsigned char buf[]);

/* Determine what index to fetch a tile number from */
static int get_tile_index(int x, int y);

/* Load a tile from a pattern table */
static void load_tile(unsigned char t, unsigned char buf[]);

/* Get a tile bitmap */
static void get_tile_bitmap(unsigned char index, unsigned char img[], int is_sprite);

/* Blit sprites */
static void blit_sprites();

/* Get a sprite pixel */
static int get_sprite_pixel(int x, int y, unsigned char *p);
static int get_sprite_pixel_8(int x, int y, unsigned char *p);
static int get_sprite_pixel_16(int x, int y, unsigned char *p);

/* Get a window pixel */
static int get_window_pixel(int x, int y, unsigned char *p);

/* Get a background pixel */
static unsigned char get_bg_pixel(int x, int y);

/* Copy a tile flipped horizontally */
static void tile_hflip(unsigned char tile[], unsigned char flipped[]);

/* Copy a tile flipped vertically */
static void tile_vflip(unsigned char tile[], unsigned char flipped[]);

/* draw a tile */
static void oam_draw_sprite(BITMAP *bmp, unsigned char tile[], int use_obp1, int x, int y);


/* Global Functions ---------------------------------------------------------*/
void lcdc_init()
{
  RGB pal[256]={0,0,0};

  LOCK_VARIABLE(lcdc_frames);

  lcdc_screen_buffer = create_bitmap(160, 144);
  clear(lcdc_screen_buffer);

  linebuffer = create_bitmap(160, 2);
  clear(linebuffer);

  sprite_img = create_bitmap(160, 144);
  clear(sprite_img);

  set_vram_source(vram);
  set_oam_source(oam);
  set_vram_handlers(vram_read, vram_write);
  set_oam_handlers(oam_read, oam_write);

  current_mode = MODE_OAMTRANSFER;
  mode_cycles = CYCLES_OAMTRANSFER;
  next_mode = MODE_DATATRANSFER;
  cycle_count = 0;

/*   shades[0] = makecol(white.r, white.g, white.b); */
/*   shades[1] = makecol(light.r, light.g, light.b); */
/*   shades[2] = makecol( dark.r,  dark.g,  dark.b); */
/*   shades[3] = makecol(black.r, black.g, black.b); */
  get_palette(pal);

  pal[shades[0]] = white;
  pal[shades[1]] = light;
  pal[shades[2]] = dark;
  pal[shades[3]] = black;

  set_palette(pal);

  lcdc_reset();
}

void lcdc_offset(int x, int y)
{
  screen_x = x;
  screen_y = y;
}

void lcdc_size(int w, int h)
{
  if ((w != 160) || (h != 144)) {
    scaled=1;
    screen_w = w;
    screen_h = h;
  }
}

void lcdc_reset()
{
  int i=0;

  for (i=0; i<CACHE_SIZE; i++) {
    if (tile_cache[i]!=NULL) {
      free(tile_cache[i]);
      tile_cache[i]=NULL;
    }
  }

  lcdc_control = 0x91;
  lcdc_y = 0;
  lcdc_status = 0;
  
  scroll_x = 0;
  scroll_y = 0;
  window_x = 0;
  window_y = 0;

  bgp = 0xFC;
  obp0 = 0xFF;
  obp1 = 0xFF;

  memset(vram, 0, 0x2000);
  memset(oam, 0, 0x9F);
  memset(&sprites[0], 0, sizeof(sprites));
}

void lcdc_cleanup()
{
  destroy_bitmap(linebuffer);
  destroy_bitmap(sprite_img);
}

unsigned char lcdc_mode()
{
  return lcdc_status & MODE_MASK;
}

unsigned char lcdc_scanline()
{
  return lcdc_y;
}

void lcdc_change(unsigned char mode)
{
  /* clear previous mode */
  lcdc_status = lcdc_status & (~MODE_MASK);
  
  /* Toggle new mode */
  lcdc_status |= mode;

  switch (mode) {
  case MODE_OAMTRANSFER:
    intr_request(INTR_LCDSTAT);
    break;
  case MODE_DATATRANSFER:
    break;
  case MODE_HBLANK:
    intr_request(INTR_LCDSTAT);
    break;
  case MODE_VBLANK:
    intr_request(INTR_VBLANK);
    break;
  }
}

/* requests an LCDC interrupt if it is enabled. */
void lcdc_interrupt(unsigned char intr)
{
}

void lcdc_trans(unsigned long cycles)
{
  if ((lcdc_control & LCD_DISPLAY) == LCD_DISPLAY) {
    cycle_count += cycles;
    switch(lcdc_status & MODE_MASK) {
    case MODE_OAMTRANSFER:
      if (cycle_count >= CYCLES_OAMTRANSFER) {
        lcdc_change(MODE_DATATRANSFER);
        cycle_count = 0;
      }
      break;
      
    case MODE_DATATRANSFER:
      if (cycle_count >= CYCLES_DATATRANSFER) {
	lcdc_change(MODE_HBLANK);
	cycle_count = 0;
      }
      break;

    case MODE_HBLANK:
      if (cycle_count >= CYCLES_HBLANK) {
	lcdc_change(MODE_OAMTRANSFER);
	lcdc_render_scanline();
	cycle_count = 0;
      }
      break;

    case MODE_VBLANK:
      if (cycle_count >= (CYCLES_VBLANK/9)) {
	lcdc_render_scanline();
	cycle_count = 0;
      }
      break;
    }
  }
}

void lcdc_lyc_compare()
{
  if (lcdc_lyc == lcdc_y) {
    lcdc_status |= 4;
    if ((lcdc_status & LYC_INTERRUPT)==LYC_INTERRUPT)
      intr_request(INTR_LCDSTAT);
  }
  else if ( (lcdc_status & 4) == 4 ) {
    lcdc_status ^= 4;
  }
}

void lcdc_render_scanline()
{
  int x, bgx, bgy;
  unsigned char p;
  
  bgx = 8;
  bgy = 7;

  if ((lcdc_control & LCD_DISPLAY)==LCD_DISPLAY) {
    lcdc_lyc_compare();
    if (lcdc_y < 144) {
      for (x=0; x<160; x++) {
	if (get_sprite_pixel(x, lcdc_y, &p)) {
	  lcdc_screen_buffer->line[lcdc_y][x] = p;
	}
	else if (get_window_pixel(x, lcdc_y, &p)) {
	  lcdc_screen_buffer->line[lcdc_y][x] = p;
	}
	else {
	  if ((lcdc_control&DISPLAY_BG)==DISPLAY_BG) {
	    lcdc_screen_buffer->line[lcdc_y][x] = get_bg_pixel(x, lcdc_y);
	  }
	  else {
	    lcdc_screen_buffer->line[lcdc_y][x] = shades[0];
	  }
	}
      }
    }

    lcdc_y += 1;
    
    if (lcdc_y == 144) {
      lcdc_refresh = 1;
      blit_sprites();
/*       lcdc_change(MODE_VBLANK); */
    }
    else if (lcdc_y == 153) {
      lcdc_y = 0;
/*       lcdc_change(MODE_OAMTRANSFER); */
    }
  }
}

void lcdc_render() 
{
/*   if ((lcdc_control&DISP_SPRITES)==DISP_SPRITES) blit_sprites(); */
  if (scaled)
    stretch_blit(lcdc_screen_buffer, lcdc_buffer, 
		 0, 0, 160, 144,                        /* Source Rectangle */
		 screen_x, screen_y, screen_w, screen_h /* Dest rectangle */
		 );
  else
    blit(lcdc_screen_buffer, lcdc_buffer, 0, 0, screen_x, screen_y, 160, 144);

  lcdc_refresh = 0;
}

/* I/O access functions */
/* Perform DMA transfer */
inline void do_dma(unsigned char addr)
{
  unsigned short staddr = addr * 0x100;
  unsigned short ctr;
  unsigned char b;

  for (ctr=0; ctr<0x9F; ++ctr) {
    b = mem_read(staddr+ctr);
    mem_write(0xFE00+ctr, b);
  }
}

/* I/O access handlers */
unsigned char lcdc_read_reg(unsigned char reg)
{
  switch(reg) {
  case 0x40:
    return lcdc_control;
  case 0x41:
    return lcdc_status;
  case 0x42:
    return scroll_y;
  case 0x43:
    return scroll_x;
  case 0x44:
    return lcdc_y;
  case 0x45:
    return lcdc_lyc;
  case 0x47:
    return bgp;
  case 0x48:
    return obp0;
  case 0x49:
    return obp1;
  case 0x4A:
    return window_x;
  case 0x4B:
    return window_y;
  default:
    return 0xFF;
  }
}

extern gbcpu_context cpu;

void lcdc_write_reg(unsigned char reg, unsigned char value)
{

  switch(reg) {
  case 0x40:
    lcdc_control = value;
    if ((lcdc_control&USE_BGMAP1)==USE_BGMAP1) {
      bgmap = &(vram[0x1C00]);
    }
    else {
      bgmap = &(vram[0x1800]);
    }

    if ((lcdc_control&WND_USE_BGMAP1)==WND_USE_BGMAP1) {
      wndmap = &(vram[0x1C00]);
    }
    else {
      wndmap = &(vram[0x1800]);
    }

    break;
  case 0x41:
    lcdc_status |= (value&(~MODE_MASK));
    break;
  case 0x42:
    scroll_y = value;
    break;
  case 0x43:
    scroll_x = value;
    break;
  case 0x44:
    lcdc_y = 0;
    lcdc_change(MODE_OAMTRANSFER);
    break;
  case 0x45:
    lcdc_lyc = value;
    break;
  case 0x46:
    do_dma(value);
    break;
  case 0x47:
    bgp = value;
    bg_pal[0] = (bgp&3);
    bg_pal[1] = ((bgp >> 2)&3);
    bg_pal[2] = ((bgp >> 4)&3);
    bg_pal[3] = ((bgp >> 6)&3);
    break;
  case 0x48:
    obp0 = value;
    ob_pal0[1] = ((obp0 >> 2)&3);
    ob_pal0[2] = ((obp0 >> 4)&3);
    ob_pal0[3] = ((obp0 >> 6)&3);
    break;
  case 0x49:
    obp1 = value;
    ob_pal1[1] = ((obp1 >> 2)&(~0xFC));
    ob_pal1[2] = ((obp1 >> 4)&(~0xFC));
    ob_pal1[3] = ((obp1 >> 6)&(~0xFC));
    break;
  case 0x4A:
    window_y = value;
    break;
  case 0x4B:
    window_x = value;
    break;
  }
}

/* Return non zero if VRAM/OAM can be accessed */
int can_access_vram()
{
  if ((lcdc_control & LCD_DISPLAY)!=LCD_DISPLAY) {
    return 1;
  }
  else {
    switch(lcdc_status & MODE_MASK) {
    case MODE_HBLANK:
      return 1;
    case MODE_VBLANK:
      return 1;
    case MODE_OAMTRANSFER:
      return 1;
    case MODE_DATATRANSFER:
      return 0;
    }
  }
}

int can_access_oam()
{
  if ((lcdc_control & LCD_DISPLAY)!=LCD_DISPLAY) {
    return 1;
  }
  else {
    switch(lcdc_status & MODE_MASK) {
    case MODE_HBLANK:
      return 1;
    case MODE_VBLANK:
      return 1;
    case MODE_OAMTRANSFER: 
      return 0;
    case MODE_DATATRANSFER:
      return 0;
    }
  }
}

void lcdc_capture(char *fname)
{  
  RGB pal[256];

  get_palette(pal);
  save_bitmap(fname, lcdc_screen_buffer, pal);

}

/* Internal function defintions ----------------------------------------------------------- */

static unsigned char vram_read(unsigned short addr)
{
  return vram[addr];
}

/* *NOTE: When one of the tile data tables is written to it invalidates corresponding
   tile cache data */
static void vram_write(unsigned short addr, unsigned char value)
{
  int tile_index;
  unsigned char packed_tile[16];

  vram[addr] = value;

  /* fix tile cache */
  if (addr <= 0x17FF) {
    tile_index = addr / 16;
    if (tile_cache[tile_index] != NULL) {      
      free(tile_cache[tile_index]);
      tile_cache[tile_index] = NULL;
    }    
  }
}

static unsigned char oam_read(unsigned short addr)
{
  return oam[addr];
}

static void oam_write(unsigned short addr, unsigned char value)
{
  unsigned char attr, lower_index, upper_index, update;
  unsigned char flipped[64];
  int index;

  oam[addr]=value;

  /* Modify sprite list */
  index = addr / 4;
  update = addr % 4;

  /* Update Y location */
  if (update==0) {
    sprites[index].y = value;
  }
  /* Update X location */
  else if (update==1) {
    sprites[index].x = value;
  }
  /* Update tile image */
  else if (update==2) {
    load_sprite(index);
  }
  /* Update attributes */
  else if (update==3) {
    load_sprite(index);
    sprites[index].use_obp1 = ( (value&USE_OBP1)==USE_OBP1 ? 1 : 0 );
  }

}

static void load_sprite(unsigned char index)
{
  unsigned char lower_index, upper_index, attr;
  unsigned char flipped[64];
  unsigned short addr = index * 4;

  attr = oam[addr+3];
  if ((lcdc_control&SPRITE16)==SPRITE16) {
    lower_index = oam[addr+2]|0x01;
    upper_index = oam[addr+2]&0xFE;

    /* Flip tile vertically if required */
    if ((attr&Y_FLIP)==Y_FLIP) {
      get_tile_bitmap(lower_index, sprites[index].upper_tile, 1);
      get_tile_bitmap(upper_index, sprites[index].lower_tile, 1);

      tile_vflip(sprites[index].upper_tile, flipped);
      memcpy(sprites[index].upper_tile, flipped, 64);

      tile_vflip(sprites[index].lower_tile, flipped);
      memcpy(sprites[index].lower_tile, flipped, 64);
    }
    else {
      get_tile_bitmap(upper_index, sprites[index].upper_tile, 1);
      get_tile_bitmap(lower_index, sprites[index].lower_tile, 1);
    }

    /* Flip the tile horizontally if required */
    if ((attr&X_FLIP)==X_FLIP) {
      tile_hflip(sprites[index].upper_tile, flipped);
      memcpy(sprites[index].upper_tile, flipped, 64);

      tile_hflip(sprites[index].lower_tile, flipped);
      memcpy(sprites[index].lower_tile, flipped, 64);
    }
  }
  else {
    upper_index = oam[addr+2];
    get_tile_bitmap(upper_index, sprites[index].upper_tile, 1);
    
    if ((attr&Y_FLIP)==Y_FLIP) {
      tile_vflip(sprites[index].upper_tile, flipped);
      memcpy(sprites[index].upper_tile, flipped, 64);
    }

    if ((attr&X_FLIP)==X_FLIP) {
      tile_hflip(sprites[index].upper_tile, flipped);
      memcpy(sprites[index].upper_tile, flipped, 64);
    }
  }

}

static void decode_tile(unsigned char tile[], unsigned char buf[])
{
  unsigned char bit, ctr, lo, hi, index=0;

  for (ctr=0; ctr<16; ctr+= 2) {
    for (bit=128; bit>0; bit=bit/2) {
      lo = ((tile[ctr]&bit)==bit);
      hi = ((tile[ctr+1]&bit)==bit);
      buf[index] = (lo | (hi << 1));
      index++;
    }
  }
}

static int get_tile_index(int x, int y)
{
  int tx, ty;

  tx = x / 8;
  ty = y / 8;

  return (ty*32)+tx;
}

static void load_tile(unsigned char t, unsigned char buf[])
{
  int index;

  if ((lcdc_control & USE_TILE0)==USE_TILE0) {
    index = t;
  }
  else {
    index = ( 0x100 + ((signed char)t) );
  }

  memcpy(buf, &(vram[index*16]), 16);
}


static void get_tile_bitmap(unsigned char index, unsigned char img[], int is_sprite)
{
  int real_index;
  char packed_tile[16];
  
  if (((lcdc_control&USE_TILE0)==USE_TILE0) || (is_sprite) ) {
    real_index = index;
  }
  else {
    real_index = ( 0x100 + ((signed char)index));
  }

  if (tile_cache[real_index]==NULL) {
    tile_cache[real_index] = malloc(64);
    memcpy(packed_tile, &(vram[real_index*16]), 16);
    decode_tile(packed_tile, (unsigned char*)tile_cache[real_index]);
  }

  memcpy(img, tile_cache[real_index], 64);

}

static void blit_sprites()
{
  int ctr, x, y;

  if ((lcdc_control&DISP_SPRITES)!=DISP_SPRITES) return;

  clear(sprite_img);
  for (ctr=0; ctr<40; ctr++) {
    x = sprites[ctr].x;
    y = sprites[ctr].y;

    if ( (x != 0) && (x <= 168) && (y != 0) && (y <= 160) ) {
      x = x - 8;
      y = y - 16;
      oam_draw_sprite(sprite_img, sprites[ctr].upper_tile,
		      sprites[ctr].use_obp1, x, y);
      
      if ((lcdc_control&SPRITE16)==SPRITE16) {
	oam_draw_sprite(sprite_img, sprites[ctr].lower_tile,
			sprites[ctr].use_obp1, x, y+8);
      }
    }
  }
}
 
static int get_sprite_pixel(int x, int y, unsigned char *p)
{
  unsigned char pixel;

  if ((lcdc_control & DISP_SPRITES)!=DISP_SPRITES) return 0;

  pixel = sprite_img->line[y][x];
  (*p) = pixel;

  if (pixel > 0)
    return 1;
  else
    return 0;
}

static int get_sprite_pixel_8(int x, int y, unsigned char *p)
{
}

static int get_sprite_pixel_16(int x, int y, unsigned char *p)
{
}


static int get_window_pixel(int x, int y, unsigned char *p) 
{
  static int prev_index=-1;
  static unsigned char tile[16]={0};
/*   static unsigned char *img; */
  static unsigned char img[64];
  int index, tx, ty, gx, gy, tile_num;
  int wx=window_x-7, wy=window_y;
  char pixel;
  unsigned char *bgmap;

  if ((lcdc_control&DISPLAY_WND)!=DISPLAY_WND) {
    (*p)=0;
    return 0;
  }

  if (!((x>=wx) && (y>=wy))) {
    (*p)=0;
    return 0;    
  }

/*   if ((lcdc_control & WND_USE_BGMAP1)==WND_USE_BGMAP1) */
/*     bgmap = &vram[0x1C00]; */
/*   else */
/*     bgmap = &vram[0x1800]; */
    
  gx = x-wx;
  gy = lcdc_y - wy;

  index = get_tile_index(gx,gy);

  if (index != prev_index) {
    prev_index = index;
    tile_num = wndmap[index];
    get_tile_bitmap(tile_num, img, FALSE);
  }
  
  tx = gx % 8;
  ty = gy % 8;
  
  pixel = img[(ty*8)+tx];
  (*p) = shades[bg_pal[pixel]];
  return 1;
}
 
static unsigned char get_bg_pixel(int x, int y)
{
  static int prev_index=0xFFF;
  static unsigned char *img;
/*   static unsigned char img[64]; */
/*   unsigned char *bgmap; */
  unsigned char *packed;
  int tx, ty, gx, gy, tile_num;
  int index;
  char pixel;

/*   if ((lcdc_control & USE_BGMAP1)==USE_BGMAP1) { */
/*     bgmap = &(vram[0x1C00]); */
/*   } */
/*   else { */
/*     bgmap = &(vram[0x1800]); */
/*   } */

  gx = x + scroll_x;
  gy = lcdc_y + scroll_y;
  if (gx >= 256)
    gx = gx - 256;
  if (gy >= 256)
    gy = gy - 256;
  
/*   index = get_tile_index(gx,gy); */
  index = ( (gy/8) * 32 ) + (gx / 8);

  if (index != prev_index) {
    prev_index = index;
    if ((lcdc_control&USE_TILE0)==USE_TILE0) {
      tile_num = bgmap[index];
    }
    else {
      tile_num = (0x100 + ((signed char)bgmap[index]));
    }

    img = tile_cache[tile_num];
    if (img == NULL) {
      img = malloc(64);
      decode_tile(&(vram[tile_num*16]), img);
      tile_cache[tile_num] = img;
    }
/*     tile_num = bgmap[index]; */
/*     get_tile_bitmap(tile_num, img, 0); */
  }

  tx = gx % 8;
  ty = gy % 8;
  
  pixel = img[(ty*8)+tx];
  return shades[bg_pal[pixel]];
}

static void tile_hflip(unsigned char tile[], unsigned char flipped[])
{
  int x, y;

  for (y=0; y<8; y++) {

    for (x=0; x<8; x++) {
      flipped[(y*8)+(7-x)] = tile[(y*8)+x];
    }

  }

}

static void tile_vflip(unsigned char tile[], unsigned char flipped[])
{
  int x, y;

  for (x=0; x<8; x++) {

    for (y=0; y<8; y++) {
      flipped[((7-y)*8)+x] = tile[(y*8)+x];
    }

  }


}

static void oam_draw_sprite(BITMAP *bmp, unsigned char tile[], int use_obp1, int x, int y)
{
  int tx, ty, ctr;
  unsigned char *pal;
  unsigned char pixel;

  if (use_obp1) 
    pal = &ob_pal1[0];
  else
    pal = &ob_pal0[0];

  tx = 0;
  ty = 0;
  for (ctr=0; ctr<64; ctr++) {
    if (tile[ctr] > 0) {
      pixel = shades[pal[tile[ctr]]];
      if ( ((x+tx) < bmp->w) && ((y+ty) < bmp->h) && 
	   ((x+tx) >= 0) && ((y+ty) >= 0) )
	bmp->line[y+ty][x+tx] = pixel;
    }

    tx++;
    if (tx==8) {
      tx = 0;
      ty++;
    }
  }
}





